import os
import csv
from ...core import SynapseProxy, Utils


class TeamAccessReport:
    """
    This report shows all the users on a team.
    """

    def __init__(self, team_ids_or_names, out_path=None):
        self._team_ids_or_names = team_ids_or_names
        if self._team_ids_or_names and not isinstance(self._team_ids_or_names, list):
            self._team_ids_or_names = [self._team_ids_or_names]
        self._out_path = Utils.expand_path(out_path) if out_path else None
        self._csv_full_path = None
        self._csv_file = None
        self._csv_writer = None
        self.errors = []

    CSV_HEADERS = ['team_id',
                   'team_name',
                   'entity_type',
                   'entity_id',
                   'entity_name',
                   'permission_level']

    def execute(self):
        raise NotImplementedError('This command has not been completed.')
        SynapseProxy.login()

        if self._out_path:
            if self._out_path.lower().endswith('.csv'):
                self._csv_full_path = self._out_path
            else:
                self._csv_full_path = os.path.join(self._out_path, 'team-access-{0}.csv'.format(Utils.timestamp_str()))
            Utils.ensure_dirs(os.path.dirname(self._csv_full_path))
            self._csv_file = open(self._csv_full_path, mode='w', newline='', encoding='utf-8')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              delimiter=',',
                                              quotechar='"',
                                              fieldnames=self.CSV_HEADERS,
                                              quoting=csv.QUOTE_ALL)
            self._csv_writer.writeheader()
        try:
            for id_or_name in self._team_ids_or_names:
                self._report_on_team(id_or_name)
        finally:
            if self._csv_file:
                self._csv_file.close()
            if self._csv_full_path:
                print('Report saved to: {0}'.format(self._csv_full_path))
        return self

    def _report_on_team(self, id_or_name):
        print('=' * 80)
        print('Looking up team: "{0}"...'.format(id_or_name))
        try:
            team = SynapseProxy.client().getTeam(id_or_name)
        except ValueError:
            # Team does not exist.
            pass
        except Exception as ex:
            self._show_error('Error loading team: {0}'.format(ex))

        if team:
            try:
                team_id = team['id']
                team_name = team['name']
                print('Team: {0} ({1})'.format(team_name, team_id))
                # TODO: How do we get the entities this team has access to?
                raise NotImplementedError()
            except Exception as ex:
                self._show_error('Error loading team data: {0}'.format(ex))
        else:
            self._show_error('Team does not exist or you do not have access to the team.')

    def _show_error(self, msg):
        self.errors.append(msg)
        Utils.eprint(msg)



import os
import csv
from ...core import SynapseProxy, Utils


class TeamMembersReport:
    """
    This report shows all the users on a team.
    """

    def __init__(self, team_ids_or_names, out_path=None):
        self._team_ids_or_names = team_ids_or_names
        if self._team_ids_or_names and not isinstance(self._team_ids_or_names, list):
            self._team_ids_or_names = [self._team_ids_or_names]
        self._out_path = Utils.expand_path(out_path) if out_path else None
        self._csv_full_path = None
        self._csv_file = None
        self._csv_writer = None
        self.errors = []

    CSV_HEADERS = ['team_id',
                   'team_name',
                   'user_id',
                   'username',
                   'first_name',
                   'last_name',
                   'company',
                   'is_admin']

    def execute(self):
        if self._out_path:
            if self._out_path.lower().endswith('.csv'):
                self._csv_full_path = self._out_path
            else:
                self._csv_full_path = os.path.join(self._out_path, 'team-members-{0}.csv'.format(Utils.timestamp_str()))
            Utils.ensure_dirs(os.path.dirname(self._csv_full_path))
            self._csv_file = open(self._csv_full_path, mode='w', newline='', encoding='utf-8')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              delimiter=',',
                                              quotechar='"',
                                              fieldnames=self.CSV_HEADERS,
                                              quoting=csv.QUOTE_ALL)
            self._csv_writer.writeheader()
        try:
            for id_or_name in self._team_ids_or_names:
                self._report_on_team(id_or_name)
        finally:
            if self._csv_file:
                self._csv_file.close()
            if self._csv_full_path:
                print('Report saved to: {0}'.format(self._csv_full_path))
        return self

    def _report_on_team(self, id_or_name):
        print('=' * 80)
        print('Looking up team: "{0}"...'.format(id_or_name))
        team = None
        try:
            team = SynapseProxy.client().getTeam(id_or_name)
        except ValueError:
            # Team does not exist.
            pass
        except Exception as ex:
            self._show_error('Error loading team: {0}'.format(ex))

        if team:
            try:
                members = list(SynapseProxy.client().getTeamMembers(team))
                print('Found team: {0} ({1}) with {2} members.'.format(team.name, team.id, len(members)))
                for record in members:
                    print('  ---')
                    member = record.get('member')
                    user = SynapseProxy.WithCache.get_user(member.get('ownerId'))

                    user_id = user.get('ownerId')
                    username = user.get('userName', None)
                    first_name = user.get('firstName', None)
                    last_name = user.get('lastName', None)
                    company = user.get('company', None)
                    is_admin = record.get('isAdmin', False)

                    if username:
                        print('  Username: {0} ({1})'.format(username, user_id))
                    if first_name:
                        print('  First Name: {0}'.format(first_name))
                    if last_name:
                        print('  Last Name: {0}'.format(last_name))
                    if company:
                        print('  Company: {0}'.format(company))
                    print('  Is Admin: {0}'.format('Yes' if is_admin else 'No'))

                    if self._csv_writer:
                        self._csv_writer.writerow({'team_id': team.id,
                                                   'team_name': team.name,
                                                   'user_id': user_id,
                                                   'username': username,
                                                   'first_name': first_name,
                                                   'last_name': last_name,
                                                   'company': company,
                                                   'is_admin': is_admin})
            except Exception as ex:
                self._show_error('Error loading team data: {0}'.format(ex))
        else:
            self._show_error('Team does not exist or you do not have access to the team.')

    def _show_error(self, msg):
        self.errors.append(msg)
        Utils.eprint(msg)



import os
import csv
from ...core import SynapseProxy, Utils


class UserTeamsReport:
    """
    This report shows all the teams a user is a member of.
    """

    def __init__(self, user_ids_or_usernames, required_member_ids_or_usernames=None, out_path=None):
        self._user_ids_or_usernames = user_ids_or_usernames
        if self._user_ids_or_usernames and not isinstance(self._user_ids_or_usernames, list):
            self._user_ids_or_usernames = [self._user_ids_or_usernames]

        self._required_member_ids_or_usernames = required_member_ids_or_usernames
        if self._required_member_ids_or_usernames and not isinstance(self._required_member_ids_or_usernames, list):
            self._required_member_ids_or_usernames = [self._required_member_ids_or_usernames]

        self._out_path = Utils.expand_path(out_path) if out_path else None
        self._csv_full_path = None
        self._csv_file = None
        self._csv_writer = None
        self.errors = []

    CSV_HEADERS = ['user_id',
                   'username',
                   'first_name',
                   'last_name',
                   'team_id',
                   'team_name',
                   'is_admin']

    def execute(self):
        SynapseProxy.login()

        if self._out_path:
            if self._out_path.lower().endswith('.csv'):
                self._csv_full_path = self._out_path
            else:
                self._csv_full_path = os.path.join(self._out_path, 'user-teams-{0}.csv'.format(Utils.timestamp_str()))
            Utils.ensure_dirs(os.path.dirname(self._csv_full_path))
            self._csv_file = open(self._csv_full_path, mode='w', newline='', encoding='utf-8')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              delimiter=',',
                                              quotechar='"',
                                              fieldnames=self.CSV_HEADERS,
                                              quoting=csv.QUOTE_ALL)
            self._csv_writer.writeheader()
        try:
            required_members = []
            required_members_usernames = []
            if self._required_member_ids_or_usernames:
                for required_user_id_or_name in self._required_member_ids_or_usernames:
                    required_user = SynapseProxy.WithCache.get_user(required_user_id_or_name)
                    if required_user is None:
                        self._show_error(
                            'Could not find user matching: {0}. Aborting.'.format(required_user_id_or_name))
                        return False
                    else:
                        required_members.append(required_user)
                        required_members_usernames.append(required_user.userName)

            if required_members:
                print('Only including teams that have members: {0}'.format(
                    ' or '.join(['{0} ({1})'.format(m.userName, m.ownerId) for m in required_members])))

            for id_or_name in self._user_ids_or_usernames:
                print('=' * 80)
                print('Looking up user: "{0}"...'.format(id_or_name))

                user = SynapseProxy.WithCache.get_user(id_or_name)

                if user:
                    user_id = user.ownerId
                    username = user.userName
                    first_name = user.get('firstName', None)
                    last_name = user.get('lastName', None)
                    print('Username: {0} ({1})'.format(username, user_id))
                    if first_name:
                        print('First Name: {0}'.format(first_name))
                    if last_name:
                        print('Last Name: {0}'.format(last_name))

                    teams = SynapseProxy.users_teams(user_id)

                    for team in teams:
                        team_id = team['id']
                        team_name = team['name']

                        # Check if the team has at least on of the required members.
                        if required_members:
                            found_match = False
                            members = SynapseProxy.WithCache.get_team_members(team_id)
                            for result in members:
                                member = result.get('member')
                                if member.get('userName') in required_members_usernames:
                                    found_match = True
                                    break

                            if not found_match:
                                continue

                        team_member = SynapseProxy.client().restGET('/team/{0}/member/{1}'.format(team_id, user_id))
                        is_admin = team_member.get('isAdmin', False)

                        print('  ---')
                        print('  Team: {0} ({1})'.format(team_name, team_id))
                        print('  Is Admin: {0}'.format('Yes' if is_admin else 'No'))
                        if self._csv_writer:
                            self._csv_writer.writerow({
                                'user_id': user_id,
                                'username': username,
                                'first_name': first_name,
                                'last_name': last_name,
                                'team_id': team_id,
                                'team_name': team_name,
                                'is_admin': is_admin
                            })
                else:
                    self._show_error('Could not find user matching: {0}'.format(id_or_name))
        finally:
            if self._csv_file:
                self._csv_file.close()
            if self._csv_full_path:
                print('Report saved to: {0}'.format(self._csv_full_path))
        return self

    def _show_error(self, msg):
        self.errors.append(msg)
        Utils.eprint(msg)



import uuid
import synapseclient as syn
from ...core import Utils, SynapseProxy
from synapseclient.core.exceptions import SynapseHTTPError


class BenefactorView(list):
    COL_BENEFACTORID = 'benefactorId'
    COL_PROJECTID = 'projectId'

    def __init__(self):
        self.scope = None
        self.view_project = None

    def set_scope(self, scope, clear=True):
        """Load a new scope object and optionally clear the benefactor list data.

        Args:
            scope: The Project, Folder, or File to scope the view to.
            clear: Whether to clear the loaded benefactor list data or not.
        """
        if clear:
            self.clear()
        self.scope = scope
        self.load()

    def load(self):
        try:
            if type(self.scope) in [syn.Project, syn.Folder, syn.File]:
                self._add_single_scope_item(self.scope)

                if type(self.scope) in [syn.Project, syn.Folder]:
                    if self.view_project is None:
                        self._create_project()

                    # Create a view and load the uniq benefactors for each folder and file in the scoped container.
                    self._create_folder_and_file_view()
            else:
                raise Exception('Scope entity must be a Project, Folder, or File.')
        except Exception as ex:
            Utils.eprint(ex)
            raise

    def _create_folder_and_file_view(self):
        """Creates a view for all the folders and files within the current scope object.
        If a view cannot be created this method will fallback to adding each folder/file individually.
        """
        try:
            self._query_view(self._create_view([syn.EntityViewType.FOLDER, syn.EntityViewType.FILE]))
        except SynapseHTTPError as ex:
            if 'scope exceeds the maximum number' in str(ex):
                print('Cannot create Folder/File view for: {0}. Falling back to individual loading and views.'.format(
                    self.scope.name))
                self._fallback_add_folders_and_files()

    def _fallback_add_folders_and_files(self):
        """This will add the benefactor data into self for each folder and file in the current scope,
        then it will try to view load each folder. If a folder cannot be view loaded it will recurse through
        this method until a folder that can be view loaded is found or each folder/file has been added individually.

        Synapse has a limit of 20,000 objects in a container. If one of the projects or folders exceeds this number
        then we need to fallback to loading the benefactor data this way.
        """
        child_items = list(SynapseProxy.client().getChildren(self.scope, includeTypes=["folder", "file"]))
        folder_ids = []

        # Manually add each folder and file in the scope that couldn't be loaded via a view.
        child_added_count = 0
        for child_item in child_items:
            child_type_name = SynapseProxy.entity_type_display_name(child_item)
            child_item_id = child_item['id']
            child_added_count += 1
            print(' - Adding {0}: {1} [{2}/{3}]'.format(child_type_name,
                                                        child_item_id,
                                                        child_added_count,
                                                        len(child_items)))
            self._add_single_scope_item(child_item_id)
            if SynapseProxy.is_folder(child_item):
                folder_ids.append(child_item_id)

        # Try to load the folders with views.
        folder_added_count = 0
        for folder_id in folder_ids:
            folder_added_count += 1
            syn_folder = SynapseProxy.client().get(folder_id)
            print(' - Creating View for Folder: {0} ({1}) [{2}/{3}]'.format(syn_folder.id,
                                                                            syn_folder.name,
                                                                            folder_added_count,
                                                                            len(folder_ids)))
            self.set_scope(syn_folder, clear=False)

    def _add_single_scope_item(self, entity_or_id):
        """Gets the benefactor data for a single entity and adds it to self.

        Args:
            entity_or_id: The entity (or entity ID) to add benefactor data for.

        Returns:
            None
        """
        entity_id = entity_or_id['id'] if isinstance(entity_or_id, syn.Entity) else entity_or_id

        paths = SynapseProxy.client().restGET('/entity/{0}/path'.format(entity_id)).get('path')
        project_id = next((p for p in paths if p['type'] == 'org.sagebionetworks.repo.model.Project'))['id']

        benefactor_id = SynapseProxy.client()._getBenefactor(entity_id).get('id')
        self._add_item(benefactor_id, project_id)

    def _query_view(self, view):
        query = 'SELECT DISTINCT {0},{1} FROM {2}'.format(self.COL_BENEFACTORID, self.COL_PROJECTID, view.id)
        query_result = SynapseProxy.client().tableQuery(query=query, resultsAs='csv')

        col_benefactorid = self._get_table_column_index(query_result.headers, self.COL_BENEFACTORID)
        col_projectid = self._get_table_column_index(query_result.headers, self.COL_PROJECTID)

        for row in query_result:
            self._add_item(row[col_benefactorid], row[col_projectid])

    def _add_item(self, benefactor_id, project_id):
        item = {
            'benefactor_id': benefactor_id,
            'project_id': project_id
        }
        if item not in self:
            self.append(item)

    def _get_table_column_index(self, headers, column_name):
        """Gets the column index for a Synapse Table Column.
        """
        for index, item in enumerate(headers):
            if item.name == column_name:
                return index

    def _create_project(self):
        name = '_TEMP_{0}_VIEW_PROJECT_'.format(str(uuid.uuid4()))
        self.view_project = SynapseProxy.client().store(syn.Project(name=name))

    def _create_view(self, entity_types):
        name = '_TEMP_{0}_VIEW_'.format(str(uuid.uuid4()))
        cols = [
            syn.Column(name=self.COL_BENEFACTORID, columnType='ENTITYID'),
            syn.Column(name=self.COL_PROJECTID, columnType='ENTITYID')
        ]
        schema = syn.EntityViewSchema(name=name,
                                      columns=cols,
                                      properties=None,
                                      parent=self.view_project,
                                      scopes=[self.scope],
                                      includeEntityTypes=entity_types,
                                      addDefaultViewColumns=False,
                                      addAnnotationColumns=False)
        return SynapseProxy.client().store(schema)

    def delete(self):
        if self.view_project:
            SynapseProxy.client().delete(self.view_project)



import functools
import os
import csv
import synapseclient as syn
from ...core import SynapseProxy, Utils


class UserProjectAccessReport:
    """
    This report will show all the Projects a user has access to.
    NOTE: Only public projects or projects the user executing this script has access to will be reported.
          This is a Synapse limitation.
    """

    def __init__(self, user_ids_or_usernames, out_path=None):
        self._user_ids_or_usernames = user_ids_or_usernames
        if self._user_ids_or_usernames and not isinstance(self._user_ids_or_usernames, list):
            self._user_ids_or_usernames = [self._user_ids_or_usernames]
        self._out_path = Utils.expand_path(out_path) if out_path else None
        self._csv_full_path = None
        self._csv_file = None
        self._csv_writer = None
        self.errors = []

    CSV_HEADERS = ['user_id',
                   'username',
                   'first_name',
                   'last_name',
                   'project_id',
                   'project_name',
                   'permission_level']

    def execute(self):
        SynapseProxy.login()

        if self._out_path:
            if self._out_path.lower().endswith('.csv'):
                self._csv_full_path = self._out_path
            else:
                self._csv_full_path = os.path.join(self._out_path,
                                                   'user-access-{0}.csv'.format(Utils.timestamp_str()))
            Utils.ensure_dirs(os.path.dirname(self._csv_full_path))
            self._csv_file = open(self._csv_full_path, mode='w', newline='', encoding='utf-8')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              delimiter=',',
                                              quotechar='"',
                                              fieldnames=self.CSV_HEADERS,
                                              quoting=csv.QUOTE_ALL)
            self._csv_writer.writeheader()
        try:
            for id_or_name in self._user_ids_or_usernames:
                print('=' * 80)
                print('Looking up user: "{0}"...'.format(id_or_name))

                try:
                    user = SynapseProxy.client().getUserProfile(id_or_name, refresh=True)
                except ValueError:
                    # User does not exist
                    user = None

                if user:
                    user_id = user.ownerId
                    username = user.userName
                    first_name = user.get('firstName', None)
                    last_name = user.get('lastName', None)
                    print('  Username: {0} ({1})'.format(username, user_id))
                    if first_name:
                        print('  First Name: {0}'.format(first_name))
                    if last_name:
                        print('  Last Name: {0}'.format(last_name))

                    for activity in SynapseProxy.users_project_access(user_id):
                        project_id = activity['id']
                        project_name = activity['name']
                        users_permission = self._get_permissions(project_id, principal_id=user_id)
                        permission_level = SynapseProxy.Permissions.name(users_permission)

                        print('    Project: {0} ({1}) [{2}]'.format(project_name, project_id, permission_level))
                        if self._csv_writer:
                            self._csv_writer.writerow({
                                'user_id': user_id,
                                'username': username,
                                'first_name': first_name,
                                'last_name': last_name,
                                'project_id': project_id,
                                'project_name': project_name,
                                'permission_level': permission_level
                            })
                else:
                    self._show_error('Could not find user matching: {0}'.format(id_or_name))
        finally:
            if self._csv_file:
                self._csv_file.close()
            if self._csv_full_path:
                print('')
                print('Report saved to: {0}'.format(self._csv_full_path))
        return self

    def _get_permissions(self, entity, principal_id):
        """Get the permissions that a user or group has on an Entity.

        :param entity:      An Entity or Synapse ID to lookup
        :param principal_id: Identifier of a user or group

        :returns: An array containing some combination of
                  ['READ', 'CREATE', 'UPDATE', 'DELETE', 'CHANGE_PERMISSIONS', 'DOWNLOAD']
                  or an empty array
        """
        # TODO: make this method return the highest permission the user has.
        principal_id = SynapseProxy.client()._getUserbyPrincipalIdOrName(principal_id)
        acl = SynapseProxy.client()._getACL(entity)

        # Look for the principal's individual permission on the entity.
        for resource in acl['resourceAccess']:
            if 'principalId' in resource and resource['principalId'] == int(principal_id):
                return resource['accessType']

        # Look for the principal's permission via a team on the entity.
        teams = self._get_users_teams(principal_id)
        team_ids = [int(t['id']) for t in teams]
        for resource in acl['resourceAccess']:
            if resource['principalId'] in team_ids:
                return resource['accessType']

        return []

    @functools.lru_cache(maxsize=SynapseProxy.WithCache.LRU_MAXSIZE, typed=True)
    def _get_users_teams(self, user_id):
        """Get all the teams a user is part of.

        Args:
            user_id: The user ID to get teams for.

        Returns:
            List
        """
        try:
            return list(SynapseProxy.users_teams(user_id))
        except syn.core.exceptions.SynapseHTTPError:
            return []

    def _show_error(self, msg):
        self.errors.append(msg)
        Utils.eprint(msg)



from .user_project_access_report import UserProjectAccessReport


def create(subparsers, parents):
    parser = subparsers.add_parser('user-project-access',
                                   parents=parents,
                                   help='Report the projects a user has access to. NOTE: Only public projects or projects the user executing this script has access to will be reported.')
    parser.add_argument('users',
                        nargs='+',
                        help='The IDs and/or usernames of the users to report on.')
    parser.add_argument('-o', '--out-path', default=None,
                        help='Path to export the report to. Specify a path that ends in ".csv" to export to a specific file otherwise a timestamped filename will be created in the out-path.')
    parser.set_defaults(_execute=execute)


def execute(args):
    return UserProjectAccessReport(
        args.users,
        out_path=args.out_path
    ).execute()



import functools
import os
import urllib
import getpass
import re
import synapseclient as syn


class SynapseProxy:
    _synapse_client = None
    _synapse_username = None
    _synapse_password = None

    @classmethod
    def configure(cls, **kwargs):
        if kwargs.get('username', None):
            cls._synapse_username = kwargs.get('username')
        if kwargs.get('password', None):
            cls._synapse_password = kwargs.get('password')

    @classmethod
    def login(cls, username=None, password=None):
        username = cls._synapse_username if username is None else username
        password = cls._synapse_password if password is None else password

        username = username or os.getenv('SYNAPSE_USERNAME')
        password = password or os.getenv('SYNAPSE_PASSWORD')

        if not username:
            username = input('Synapse username: ')

        if not password:
            password = getpass.getpass(prompt='Synapse password: ')

        print('Logging into Synapse as: {0}'.format(username))
        try:
            # Disable the synapseclient progress output.
            syn.core.utils.printTransferProgress = lambda *a, **k: None

            cls._synapse_client = syn.Synapse(skip_checks=True)
            cls._synapse_client.login(username, password, silent=True, rememberMe=False)
        except Exception as ex:
            cls._synapse_client = None
            print('Synapse login failed: {0}'.format(str(ex)))

        return cls._synapse_client is not None

    @classmethod
    def logged_in(cls):
        return cls._synapse_client is not None

    @classmethod
    def client(cls):
        if not cls._synapse_client:
            cls.login()
        return cls._synapse_client

    @classmethod
    def is_synapse_id(cls, value):
        """Gets if a string is a Synapse ID.

        Args:
            value: The string to check.

        Returns:
            True or False.
        """
        if isinstance(value, str):
            return re.match('^syn[0-9]+$', value.strip(), re.IGNORECASE) is not None
        else:
            return False

    PROJECT_TYPE_DISPLAY_NAME = 'Project'
    FOLDER_TYPE_DISPLAY_NAME = 'Folder'
    FILE_TYPE_DISPLAY_NAME = 'File'
    LINK_TYPE_DISPLAY_NAME = 'Line'
    TABLE_TYPE_DISPLAY_NAME = 'Table'

    CONCRETE_TYPE_MAPPING = {
        'org.sagebionetworks.repo.model.FileEntity': FILE_TYPE_DISPLAY_NAME,
        'org.sagebionetworks.repo.model.Folder': FOLDER_TYPE_DISPLAY_NAME,
        'org.sagebionetworks.repo.model.Link': LINK_TYPE_DISPLAY_NAME,
        'org.sagebionetworks.repo.model.Project': PROJECT_TYPE_DISPLAY_NAME,
        'org.sagebionetworks.repo.model.table.TableEntity': TABLE_TYPE_DISPLAY_NAME
    }

    @classmethod
    def _extract_concrete_type(cls, obj):
        result = obj
        if not isinstance(obj, str):
            for key in ['concreteType', 'type']:
                if key in obj:
                    result = obj[key]
                    break

        if str(result) not in cls.CONCRETE_TYPE_MAPPING.keys():
            raise ValueError('Cannot extract type from: {0}'.format(obj))

        return result

    @classmethod
    def entity_type_display_name(cls, concrete_type):
        concrete_type = cls._extract_concrete_type(concrete_type)
        return cls.CONCRETE_TYPE_MAPPING.get(concrete_type, 'Unknown Type: {0}'.format(concrete_type))

    @classmethod
    def is_project(cls, concrete_type):
        return cls.entity_type_display_name(concrete_type) == cls.PROJECT_TYPE_DISPLAY_NAME

    @classmethod
    def is_folder(cls, concrete_type):
        return cls.entity_type_display_name(concrete_type) == cls.FOLDER_TYPE_DISPLAY_NAME

    @classmethod
    def is_file(cls, concrete_type):
        return cls.entity_type_display_name(concrete_type) == cls.FILE_TYPE_DISPLAY_NAME

    @classmethod
    def users_teams(cls, user_id):
        """Gets all the teams a user is part of.

        Args:
            user_id:

        Returns:

        """
        for item in cls.client()._GET_paginated('/user/{0}/team/'.format(user_id)):
            yield item

    @classmethod
    def users_project_access(cls, user_id, **kwparams):
        """ Gets the Projects a user has access to.

        https://rest-docs.synapse.org/rest/GET/projects/user/principalId.html

        Args:
            user_id: The user ID to get activity for.
            **kwparams: Params for the GET request.

        Returns:
            Generator
        """
        request = (kwparams or {})

        response = {"nextPageToken": "first"}
        while response.get('nextPageToken') is not None:
            url_params = urllib.parse.urlencode(request)
            uri = '/projects/user/{0}?{1}'.format(user_id, url_params)

            response = cls.client().restGET(uri)
            for child in response['results']:
                yield child
            request['nextPageToken'] = response.get('nextPageToken', None)

    class WithCache:
        LRU_MAXSIZE = (os.cpu_count() or 1) * 16

        @classmethod
        @functools.lru_cache(maxsize=LRU_MAXSIZE, typed=True)
        def get_user(cls, username_or_id):
            try:
                return SynapseProxy.client().getUserProfile(username_or_id, refresh=True)
            except (ValueError, syn.core.exceptions.SynapseHTTPError):
                return None

        @classmethod
        @functools.lru_cache(maxsize=LRU_MAXSIZE, typed=True)
        def get_team(cls, team_id_or_name):
            try:
                return SynapseProxy.client().getTeam(team_id_or_name)
            except (ValueError, syn.core.exceptions.SynapseHTTPError):
                return None

        @classmethod
        @functools.lru_cache(maxsize=LRU_MAXSIZE, typed=True)
        def get_user_or_team(cls, user_id_or_team_id):
            # NOTE: User and Team IDs do NOT overlap in Synapse.
            return SynapseProxy.WithCache.get_user(user_id_or_team_id) or \
                   SynapseProxy.WithCache.get_team(user_id_or_team_id)

        @classmethod
        @functools.lru_cache(maxsize=LRU_MAXSIZE, typed=True)
        def get_team_members(cls, team_id):
            try:
                return list(SynapseProxy.client().getTeamMembers(team_id))
            except (ValueError, syn.core.exceptions.SynapseHTTPError):
                return []

    class Permissions:
        ADMIN = [
            'UPDATE',
            'DELETE',
            'CHANGE_PERMISSIONS',
            'CHANGE_SETTINGS',
            'CREATE',
            'DOWNLOAD',
            'READ',
            'MODERATE'
        ]

        CAN_EDIT_AND_DELETE = [
            'DOWNLOAD',
            'UPDATE',
            'CREATE',
            'DELETE',
            'READ'
        ]

        CAN_EDIT = [
            'DOWNLOAD',
            'UPDATE',
            'CREATE',
            'READ'
        ]

        CAN_DOWNLOAD = [
            'DOWNLOAD',
            'READ'
        ]

        CAN_VIEW = [
            'READ'
        ]

        TEAM_MANAGER = [
            'SEND_MESSAGE',
            'READ',
            'UPDATE',
            'TEAM_MEMBERSHIP_UPDATE',
            'DELETE'
        ]

        @classmethod
        def name(cls, codes):
            codes = set(codes)
            if set(cls.ADMIN) == codes:
                return 'Administrator'
            elif set(cls.CAN_DOWNLOAD) == codes:
                return 'Can download'
            elif set(cls.CAN_VIEW) == codes:
                return 'Can view'
            elif set(cls.CAN_EDIT) == codes:
                return 'Can edit'
            elif set(cls.CAN_EDIT_AND_DELETE) == codes:
                return 'Can edit & delete'



from .entity_permissions_report import EntityPermissionsReport


def create(subparsers, parents):
    parser = subparsers.add_parser('entity-permissions',
                                   parents=parents,
                                   help='Report the permissions of each user and team on a Synapse entity.')
    parser.add_argument('entities',
                        nargs='+',
                        help='The IDs and/or names of the entities to report on.')
    parser.add_argument('-o', '--out-path', default=None,
                        help='Path to export the report to. Specify a path that ends in ".csv" to export to a specific file otherwise a timestamped filename will be created in the out-path.')
    parser.add_argument('-r', '--recursive',
                        default=False,
                        action='store_true',
                        help='Recursively report permissions on child entities. Will report on each sub-folder/file/table that has different permissions from the starting entity.')
    parser.add_argument('-a', '--all',
                        default=False,
                        action='store_true',
                        help='Report permissions on every entity regardless of the parent permission.')
    parser.set_defaults(_execute=execute)


def execute(args):
    return EntityPermissionsReport(
        args.entities,
        out_path=args.out_path,
        recursive=args.recursive,
        report_on_all=args.all
    ).execute()



from .user_teams_report import UserTeamsReport


def create(subparsers, parents):
    parser = subparsers.add_parser('user-teams',
                                   parents=parents,
                                   help='Report the teams a user is a member of.')
    parser.add_argument('users',
                        nargs='+',
                        help='The IDs and/or usernames of the users to report on.')
    parser.add_argument('-o', '--out-path', default=None,
                        help='Path to export the report to. Specify a path that ends in ".csv" to export to a specific file otherwise a timestamped filename will be created in the out path.')
    parser.add_argument('--has-member',
                        help='Only report teams that also have this user in the team.',
                        action='append',
                        nargs='?')
    parser.set_defaults(_execute=execute)


def execute(args):
    return UserTeamsReport(
        args.users,
        required_member_ids_or_usernames=args.has_member,
        out_path=args.out_path
    ).execute()



import os
import sys
import datetime


class Utils:
    @staticmethod
    def eprint(*args, **kwargs):
        """Print to stderr"""
        print(*args, file=sys.stderr, **kwargs)

    @staticmethod
    def expand_path(local_path):
        var_path = os.path.expandvars(local_path)
        expanded_path = os.path.expanduser(var_path)
        return os.path.abspath(expanded_path)

    @staticmethod
    def ensure_dirs(local_path):
        """Ensures the directories in local_path exist.

        Args:
            local_path: The local path to ensure.

        Returns:
            None
        """
        if not os.path.isdir(local_path):
            os.makedirs(local_path)

    @staticmethod
    def timestamp_str():
        return datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")



from .team_access_report import TeamAccessReport


def create(subparsers, parents):
    parser = subparsers.add_parser('team-access',
                                   parents=parents,
                                   help='Report the entities a team has access to.')
    parser.add_argument('teams',
                        nargs='+',
                        help='The IDs and/or names of the teams to report on.')
    parser.add_argument('-o', '--out-path', default=None,
                        help='Path to export the report to. Specify a path that ends in ".csv" to export to a specific file otherwise a timestamped filename will be created in the out path.')
    parser.set_defaults(_execute=execute)


def execute(args):
    return TeamAccessReport(
        args.teams,
        out_path=args.out_path
    ).execute()



import argparse
import sys

from .commands.team_members_report import cli as team_members_report_cli
from .commands.benefactor_permissions_report import cli as benefactor_permissions_report_cli
from .commands.entity_permissions_report import cli as entity_permissions_report_cli
from .commands.user_project_access_report import cli as user_project_access_report_cli
from .commands.user_teams_report import cli as user_teams_report_cli
from .commands.team_access_report import cli as team_access_report_cli  # TODO: Uncomment when fully implemented.
from .core import SynapseProxy
from ._version import __version__

ALL_ACTIONS = [
    benefactor_permissions_report_cli,
    entity_permissions_report_cli,
    user_project_access_report_cli,
    user_teams_report_cli,
    team_members_report_cli
]


def main(args=None):
    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument('-u', '--username', help='Synapse username.', default=None)
    shared_parser.add_argument('-p', '--password', help='Synapse password.', default=None)

    main_parser = argparse.ArgumentParser(description='Synapse Reports')
    main_parser.add_argument('--version', action='version', version='%(prog)s {0}'.format(__version__))
    subparsers = main_parser.add_subparsers(title='Commands', dest='command')
    for action in ALL_ACTIONS:
        action.create(subparsers, [shared_parser])

    cmd_args = main_parser.parse_args(args)

    if '_execute' in cmd_args:
        try:
            SynapseProxy.configure(username=cmd_args.username, password=cmd_args.password)
            cmd = cmd_args._execute(cmd_args)
            if cmd.errors:
                print('Finished with errors.')
                for error in cmd.errors:
                    print(error)
                sys.exit(1)
            else:
                print('Finished successfully.')
                sys.exit(0)
        except Exception as ex:
            print(ex)
            sys.exit(1)
    else:
        main_parser.print_help()
        sys.exit(1)



import setuptools
from src.syn_reports._version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="syn-reports",
    version=__version__,
    author="Patrick Stout",
    author_email="pstout@prevagroup.com",
    license="Apache2",
    description="Synapse reporting.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ki-tools/syn-reports",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=(
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
    entry_points={
        'console_scripts': [
            "syn-reports = syn_reports.cli:main"
        ]
    },
    install_requires=[
        "synapseclient>=2.1.0,<3.0.0"
    ]
)



import os
import csv
import synapseclient as syn
from ...core import SynapseProxy, Utils


class EntityPermissionsReport:
    """
    This report will show the permissions of each user and team on an entity.
    """

    def __init__(self, entity_ids_or_names, out_path=None, recursive=False, report_on_all=False):
        self._entity_ids_or_names = entity_ids_or_names
        if self._entity_ids_or_names and not isinstance(self._entity_ids_or_names, list):
            self._entity_ids_or_names = [self._entity_ids_or_names]
        self._out_path = Utils.expand_path(out_path) if out_path else None
        self._recursive = recursive
        self._report_on_all = report_on_all
        self._csv_full_path = None
        self._csv_file = None
        self._csv_writer = None
        self.errors = []

    CSV_HEADERS = ['entity_type',
                   'entity_id',
                   'entity_name',
                   'principal_type',
                   'team_id',
                   'team_name',
                   'user_id',
                   'username',
                   'first_name',
                   'last_name',
                   'emails',
                   'permission_level']

    def execute(self):
        if self._out_path:
            if self._out_path.lower().endswith('.csv'):
                self._csv_full_path = self._out_path
            else:
                self._csv_full_path = os.path.join(self._out_path,
                                                   'entity-permissions-{0}.csv'.format(Utils.timestamp_str()))
            Utils.ensure_dirs(os.path.dirname(self._csv_full_path))
            self._csv_file = open(self._csv_full_path, mode='w', newline='', encoding='utf-8')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              delimiter=',',
                                              quotechar='"',
                                              fieldnames=self.CSV_HEADERS,
                                              quoting=csv.QUOTE_ALL)
            self._csv_writer.writeheader()
        try:
            for id_or_name in self._entity_ids_or_names:
                self._report_on_entity(id_or_name)
        finally:
            if self._csv_file:
                self._csv_file.close()
            if self._csv_full_path:
                print('')
                print('Report saved to: {0}'.format(self._csv_full_path))
        return self

    def _report_on_entity(self, id_or_name, root_benefactor_id=None):
        print('=' * 80)
        print('Looking up entity: "{0}"...'.format(id_or_name))
        entity = None
        try:
            if SynapseProxy.is_synapse_id(id_or_name):
                syn_id_to_load = id_or_name
            else:
                syn_id_to_load = SynapseProxy.client().findEntityId(id_or_name)

            if syn_id_to_load:
                entity = SynapseProxy.client().restGET('/entity/{0}/type'.format(syn_id_to_load))
        except syn.core.exceptions.SynapseHTTPError:
            # Entity does not exist.
            pass
        except Exception as ex:
            self._show_error('Error loading entity: {0}'.format(ex))

        if entity:
            try:
                entity_type = SynapseProxy.entity_type_display_name(entity)
                print('{0}: {1} ({2}) found.'.format(entity_type, entity['name'], entity['id']))
                benefactor_id = entity['benefactorId']

                # Only report on permissions that are different from the root entity's permissions.
                if not self._report_on_all and (root_benefactor_id is not None and root_benefactor_id == benefactor_id):
                    print('  Permissions inherited from root entity.')
                else:
                    # NOTE: Do not use syn._getACL() as it will raise an error if the entity inherits its ACL.
                    entity_acl = SynapseProxy.client().restGET('/entity/{0}/acl'.format(benefactor_id))
                    # Get the resource access items and sort them so they can be compared.
                    resource_accesses = sorted(entity_acl.get('resourceAccess', []), key=lambda r: r.get('principalId'))
                    for resource in resource_accesses:
                        resource.get('accessType').sort()

                    for resource in resource_accesses:
                        user_or_team = SynapseProxy.WithCache.get_user_or_team(resource.get('principalId'))
                        permission_level = SynapseProxy.Permissions.name(resource.get('accessType'))
                        self._display_principal(entity, entity_type, permission_level, user_or_team)

                        if isinstance(user_or_team, syn.Team):
                            members = SynapseProxy.WithCache.get_team_members(user_or_team.id)
                            for record in members:
                                member = record.get('member')
                                user_id = member.get('ownerId')
                                user = SynapseProxy.WithCache.get_user(user_id)
                                self._display_principal(entity,
                                                        entity_type,
                                                        permission_level,
                                                        user,
                                                        from_team_id=user_or_team.id,
                                                        from_team_name=user_or_team.name)

                if self._recursive:
                    if root_benefactor_id is None:
                        root_benefactor_id = benefactor_id

                    if SynapseProxy.is_project(entity) or SynapseProxy.is_folder(entity):
                        for child in SynapseProxy.client().getChildren(entity['id'],
                                                                       includeTypes=['folder', 'file', 'table']):
                            self._report_on_entity(child['id'], root_benefactor_id=root_benefactor_id)

            except Exception as ex:
                self._show_error('Error loading entity data: {0}'.format(ex))
        else:
            self._show_error('Entity does not exist or you do not have access to the entity.')

    def _display_principal(self, entity, entity_type, permission_level, user_or_team,
                           from_team_id=None, from_team_name=None):
        indent = '  ' if from_team_id is None else '    '
        print('{0}---'.format(indent))
        principal_type = None
        team_name = None
        team_id = None
        user_id = None
        username = None
        first_name = None
        last_name = None
        emails = None

        if isinstance(user_or_team, syn.Team):
            principal_type = 'Team'
            team_name = user_or_team.name
            team_id = user_or_team.id
            print('{0}Team: {1} ({2})'.format(indent, team_name, team_id))
        elif user_or_team is None:
            principal_type = 'Unknown'
            print('{0}Username/Team: Unknown - Script user ({1}) may not have access to this user/team data.)'.format(
                indent, SynapseProxy._synapse_username))
            if from_team_name:
                print('{0}From Team: {1} ({2})'.format(indent, from_team_name, from_team_id))
        else:
            principal_type = 'User'
            user_id = user_or_team.ownerId
            username = user_or_team.userName
            first_name = user_or_team.get('firstName', None)
            last_name = user_or_team.get('lastName', None)
            emails = ','.join(user_or_team.get('emails', []))
            print('{0}Username: {1} ({2})'.format(indent, username, user_id))
            if from_team_name:
                print('{0}From Team: {1} ({2})'.format(indent, from_team_name, from_team_id))
            if first_name:
                print('{0}First Name: {1}'.format(indent, first_name))
            if last_name:
                print('{0}Last Name: {1}'.format(indent, last_name))
            if emails:
                print('{0}Emails: {1}'.format(indent, emails))

        print('{0}Permission: {1}'.format(indent, permission_level))

        if self._csv_writer:
            self._csv_writer.writerow({
                'entity_type': entity_type,
                'entity_id': entity['id'],
                'entity_name': entity['name'],
                'principal_type': principal_type,
                'team_id': team_id or from_team_id,
                'team_name': team_name or from_team_name,
                'user_id': user_id,
                'username': username,
                'first_name': first_name,
                'last_name': last_name,
                'emails': emails,
                'permission_level': permission_level
            })

    def _show_error(self, msg):
        self.errors.append(msg)
        Utils.eprint(msg)
